import json
import pathlib


class Config:
    def __init__(self,
                 root_dir='.',
                 projects=[],
                 **kwargs):
        self.root_dir = root_dir
        self.projects = projects

    @classmethod
    def from_file(cls, realm_json_file):
        p = pathlib.Path(realm_json_file)
        with open(realm_json_file, 'r') as f:
            cfg_json = json.load(f)

            return cls(root_dir=str(p.parent),
                       **cfg_json)



from typing import Type

import click

from .base_command import BaseCommand


class Group(click.Group):
    def add_command(self, cmd: Type[BaseCommand], name=None):
        created_cmd = cmd.to_command()
        if not created_cmd.params:
            created_cmd.params = []
        created_cmd.params.extend(self.params)
        super().add_command(created_cmd,
                            name=name)

    def add_group(self, grp: click.Group, name=None):
        if not grp.params:
            grp.params = []
        grp.params.extend(self.params)
        # Call the original add_command that also handles group
        super(Group, self).add_command(grp,
                                       name=name)



import shlex

import subprocess


class ChildProcess:
    FORCE_CAPTURE = False
    CAPTURE_PARAMS = dict(stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)

    @classmethod
    def run(cls, command, **kwargs) -> str:
        params = dict(cls.CAPTURE_PARAMS)
        params.update(kwargs)

        if cls.FORCE_CAPTURE:
            params.update(cls.CAPTURE_PARAMS)

        if not kwargs.get('shell'):
            command = shlex.split(command, posix=False)
        p = subprocess.Popen(command,
                             **params)
        out, err = p.communicate()
        if out and isinstance(out, bytes):
            out = out.decode()

        if err and isinstance(err, bytes):
            err = err.decode()

        if p.returncode != 0:
            raise RuntimeError(f'Failed running command: {command};\nstdout: {out};\nstderr: {err};')

        return out



from typing import List

from .config import Config
from .project import Project


class RealmContext:
    def __init__(self,
                 config: Config,
                 projects: List[Project]):
        self.config = config
        self.projects = projects
        self.all_projects = projects



import glob
import json
import os
import pathlib
import sys
from typing import List

import click
from realm.entities import Config, Project, RealmContext
from realm.version import __version__

from .commands.init import InitCommand
from .commands.install import InstallCommand
from .commands.ls import LsCommand
from .commands.run import RunCommand
from .commands.task import TaskCommand
from .core.params import GlobalOption
from .realm_click_types import RealmClickGroup

CONFIG_FILE = 'realm.json'


class Application:
    @classmethod
    def create(cls):
        grp = RealmClickGroup(callback=cls.init_context,
                              params=cls.global_options())

        grp.add_command(InitCommand)
        grp.add_command(LsCommand)
        grp.add_command(RunCommand)
        grp.add_command(InstallCommand)
        grp.add_command(TaskCommand)

        return grp

    @classmethod
    def global_options(cls):
        return [
            GlobalOption(['--version', '-V'],
                         is_flag=True,
                         callback=cls.print_version,
                         help='Display realm version'),
            GlobalOption(['--parallelism', '-p'],
                         type=click.INT,
                         show_default=True,
                         default=1,
                         help='Sets the parallelism for the command (if supported)'),
            GlobalOption(['--since'],
                         help='Includes only projects changed since the specified ref'),
            GlobalOption(['--all'],
                         is_flag=True,
                         help='Include all projects if no projects were changed when using the --since filter'),
            GlobalOption(['--scope'],
                         type=click.STRING,
                         help='Includes only projects that match the given pattern'),
            GlobalOption(['--ignore'],
                         type=click.STRING,
                         help='Filters out projects that match the given pattern'),
            GlobalOption(['--match'],
                         type=click.STRING,
                         help='Filters by a field specified in `pyproject.toml`')
        ]

    @staticmethod
    def print_version(ctx, value):
        if value:
            msg = 'Realm {}'.format(click.style(__version__, fg='yellow'))
            click.echo(msg)

            sys.exit(0)

    @staticmethod
    @click.pass_context
    def init_context(ctx, **kwargs):
        cfg = Application.read_config()
        projects = Application.get_projects(cfg)
        ctx.obj = RealmContext(config=cfg,
                               projects=projects)

    @classmethod
    def read_config(cls):
        current = pathlib.Path(os.getcwd())
        current = current.joinpath(CONFIG_FILE)
        possible_paths = [x.joinpath(CONFIG_FILE) for x in current.parents]
        for p in possible_paths:
            if p.exists():
                with p.open() as f:
                    cfg_json = json.load(f)
                    return Config(root_dir=str(p.parent),
                                  **cfg_json)

        return Config()

    @classmethod
    def get_projects(cls, cfg: Config) -> List[Project]:
        tmp_candidates = [glob.glob(os.path.join(cfg.root_dir, p)) for p in cfg.projects]
        candidates = []
        for lst in tmp_candidates:
            for c in lst:
                candidates.append(os.path.abspath(c))

        projects = []
        for path in candidates:
            if 'pyproject.toml' in os.listdir(path):
                projects.append(Project(source_dir=path,
                                        root_dir=cfg.root_dir))

        return projects


cli = Application.create()



import os
from pathlib import Path

import toml
from realm.utils.child_process import ChildProcess

PYPROJECT_FILE = 'pyproject.toml'


class Project:
    def __init__(self, source_dir, root_dir):
        self.source_dir = Path(source_dir)
        self.name = os.path.basename(source_dir)
        self.relative_path = source_dir[len(root_dir):].lstrip(os.sep)

        toml_path = str(self.source_dir.joinpath(PYPROJECT_FILE))
        self.pyproject = toml.load(toml_path)

    @property
    def version(self) -> str:
        return self.pyproject['tool']['poetry']['version']

    @property
    def package_name(self) -> str:
        return self.pyproject['tool']['poetry']['name']

    def extract_field(self, toml_path: str):
        parts = toml_path.split('.')
        current = self.pyproject
        for p in parts:
            current = current.get(p, {})
        if bool(current):
            return current
        return None

    @property
    def dependencies(self):
        all_dependencies = dict(self.pyproject['tool']['poetry']['dev-dependencies'])
        all_dependencies.update(self.pyproject['tool']['poetry']['dependencies'])

        return all_dependencies

    def has_task(self, task_name) -> bool:
        tasks = self.pyproject['tool'].get('poe', {}).get('tasks', {})
        return task_name in tasks

    def execute_cmd(self, cmd, **kwargs):
        full_cmd = cmd
        env = self._create_env()

        try:
            params = dict(stdout=None,
                          stderr=None,
                          shell=True,
                          env=env,
                          cwd=self.source_dir)
            params.update(kwargs)
            return ChildProcess.run(full_cmd,
                                    **params)
        except RuntimeError as e:
            msg = f'{str(e)}\nproject: {os.path.basename(self.source_dir)}'
            raise RuntimeError(msg)

    def __repr__(self):
        return self.package_name

    def _create_env(self):
        env = dict(os.environ)
        current_venv = os.getenv('VIRTUAL_ENV', os.getenv('CONDA_PREFIX'))
        if current_venv:
            path_env = os.environ.get('PATH', '')
            # Remove venv from path
            env['PATH'] = ':'.join([e for e
                                    in path_env.split(':')
                                    if current_venv not in e])
            env.pop('VIRTUAL_ENV', None)
            env.pop('CONDA_PREFIX', None)
        env['REALM_PROJECT_NAME'] = self.name
        env['REALM_PROJECT_PATH'] = str(self.source_dir)

        return env



import click
from realm.cli.realm_command import RealmCommand
from realm.utils import await_all


class TaskCommand(RealmCommand[dict]):
    NAME = 'task'
    HELP_MESSAGE = """
    Runs a poe task on all projects containing this task
    """
    PARAMS = [
        click.Argument(['task_name'], type=click.STRING)
    ]

    def _filter_projects(self):
        super()._filter_projects()
        task_name = self.params['task_name']
        self.ctx.projects = [p for p
                             in self.ctx.projects
                             if p.has_task(task_name)]

    def run(self):
        task_name = self.params['task_name']
        failed_projects = []
        futures = []
        for project in self.ctx.projects:
            f = self.pool.submit(self.execute_task, project, task_name, failed_projects)
            futures.append(f)

        await_all(futures)

        if any(failed_projects):
            names = [p.name for p in failed_projects]
            formatted_names = ', '.join(names)
            raise RuntimeError(f'Task {task_name} failed for projects: {formatted_names}')

    @staticmethod
    def execute_task(project, task_name, failed_projects):
        click.echo(f'Running task {task_name} for: {project.name}')
        try:
            out = project.execute_cmd(f'poetry run poe {task_name}')
            if out:
                click.echo(out)
        except RuntimeError as e:
            click.echo(e, err=True)
            failed_projects.append(project)



import click
from realm.cli.realm_command import RealmCommand


class LsCommand(RealmCommand[dict]):
    NAME = 'ls'
    HELP_MESSAGE = """
    Prints all projects managed
    """
    PARAMS = [
        click.Option(['--paths'],
                     help='Prints relative paths of the projects',
                     is_flag=True)
    ]

    def run(self):
        for p in self.ctx.projects:
            if self.params.get('paths'):
                click.echo(f'{p.relative_path}')
            else:
                click.echo(f'{p.name}@{p.version}')



import click
from realm.cli.realm_command import RealmCommand
from realm.utils import await_all


class RunCommand(RealmCommand[dict]):
    NAME = 'run'
    HELP_MESSAGE = """
    Executes a command on all projects
    """
    PARAMS = [
        click.Argument(['command'], type=click.STRING, nargs=-1)
    ]

    def run(self):
        cmd = self.params.get('command')
        full_cmd = ' '.join(cmd)
        futures = [self.pool.submit(project.execute_cmd, full_cmd)
                   for project
                   in self.ctx.projects]

        await_all(futures)



import click

from .core.params import GlobalOption


class RealmFormatter(click.HelpFormatter):
    def __init__(self, headers_color=None, options_color=None, *args, **kwargs):
        self.headers_color = headers_color
        self.options_color = options_color
        super().__init__(*args, **kwargs)

    def write_usage(self, prog, args='', prefix='Usage: '):
        colorized_prefix = click.style(prefix, fg=self.headers_color)
        super().write_usage(prog,
                            args,
                            prefix=colorized_prefix)

    def write_heading(self, heading):
        self.write(click.style('', fg=self.headers_color, reset=False))
        super().write_heading(heading)
        self.write(click.style('', reset=True))

    def write_dl(self, rows, **kwargs):
        colorized_rows = [(click.style(row[0], fg=self.options_color), row[1])
                          for row
                          in rows]
        super().write_dl(colorized_rows, **kwargs)


class RealmFormatHelpMixin(click.Command):
    def get_help(self, ctx):
        formatter = RealmFormatter(
            width=500,
            max_width=ctx.max_content_width,
            headers_color='yellow',
            options_color='bright_blue')
        self.format_help(ctx, formatter)
        return formatter.getvalue().rstrip('\n')

    def format_options(self, ctx, formatter):
        """Writes all the options into the formatter if they exist."""
        opts = []
        global_opts = []
        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            if rv is not None:
                if self.__is_global_option(param):
                    global_opts.append(rv)
                else:
                    opts.append(rv)

        if global_opts:
            with formatter.section("Global Options"):
                formatter.write_dl(global_opts)

        if opts:
            with formatter.section("Options"):
                formatter.write_dl(opts)

    @staticmethod
    def __is_global_option(opt: click.Option):
        return isinstance(opt, GlobalOption) or opt.name == 'help'



import fnmatch
from concurrent.futures import ThreadPoolExecutor, Executor
import os
from abc import ABC

from realm.entities import RealmContext

from .core.base_command import BaseCommand, T
from .realm_click_types import RealmClickCommand
from ..utils.filters import apply_since_filters


class RealmCommand(BaseCommand[T], ABC):
    CLICK_COMMAND_CLS = RealmClickCommand

    def __init__(self, ctx: RealmContext, **kwargs):
        super().__init__(ctx, **kwargs)
        # Just for the type annotation
        self.ctx = ctx
        self.pool = ThreadPoolExecutor(self._params.get('parallelism', 1))
        self._filter_projects()
        os.environ['REALM_ROOT_PATH'] = self.ctx.config.root_dir

    def _filter_projects(self):
        # TODO: refactor into smaller methods, allow multiple values
        since = self._params.get('since')
        include_all_when_empty = self._params.get('all')
        if since:
            apply_since_filters(self.ctx, since)
            if include_all_when_empty:
                if not any(self.ctx.projects):
                    self.ctx.projects = self.ctx.all_projects

        scope = self._params.get('scope')
        if scope:
            self.ctx.projects = [p for p
                                 in self.ctx.projects
                                 if fnmatch.fnmatch(p.name, scope)]

        ignore = self._params.get('ignore')
        if ignore:
            self.ctx.projects = [p for p
                                 in self.ctx.projects
                                 if not fnmatch.fnmatch(p.name, scope)]

        match = self._params.get('match')
        if match:
            field, _, value = match.partition('=')
            f1 = f'tool.realm.{field}'
            f2 = f'tool.{field}'
            self.ctx.projects = [p for p
                                 in self.ctx.projects
                                 if str(p.extract_field(f1)) == value or
                                 str(p.extract_field(f2)) == value]



import click
from realm.cli.realm_command import RealmCommand
from realm.utils import await_all


class InstallCommand(RealmCommand[dict]):
    NAME = 'install'
    HELP_MESSAGE = """
    Executes "poetry install" on all projects
    """

    def run(self):
        futures = [self.pool.submit(self._install, project)
                   for project
                   in self.ctx.projects]

        await_all(futures)

    @staticmethod
    def _install(project):
        out = project.execute_cmd('poetry install')
        if out:
            # used only for tests :(
            click.echo(out)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['realm',
 'realm.cli',
 'realm.cli.commands',
 'realm.cli.core',
 'realm.entities',
 'realm.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0.0,<8.0.0', 'colorama>=0.4.4,<0.5.0', 'toml>=0.10.2,<0.11.0']

extras_require = \
{'tasks': ['poethepoet>=0.9.0,<0.10.0']}

entry_points = \
{'console_scripts': ['realm = realm.cli.application:cli']}

setup_kwargs = {
    'name': 'realm',
    'version': '0.1.0a1',
    'description': 'A tool for managing python poetry projects',
    'long_description': '## Realm\nRealm is tool for managing multiple poetry projects in the same git repository.\n\nThis project is inspired by the `lerna` project available for JavaScript\n\n[![Build Status](https://github.com/orlevii/realm/workflows/build/badge.svg?branch=master&event=push)](https://github.com/orlevii/realm/actions/workflows/build.yml?query=branch%3Amaster)\n\n### Requirements\nIn order to start using realm, you first need to have [poetry](https://github.com/python-poetry/poetry) installed\n\n### Commands\n* <code>realm init</code> - Initializes a new realm repo\n* <code>realm install</code> - Executes `poetry install` on all projects\n* <code>realm ls</code> - Prints all projects managed\n* <code>realm run</code> - Executes a command on all projects\n* <code>realm task</code> - Runs a poe task on all projects containing this task (requires poethepoet)\n\n#### Filtering\nYou can set up filters to affect only certain projects\n\nFor example, you can install only changed projects \n```bash\n$ realm install --since origin/master\n```\n\nAvailable filters:\n* <code>--since</code> - Includes only projects changed since the specified ref\n* <code>--scope</code> - Includes only projects that match the given pattern\n* <code>--ignore</code> - Filters out projects that match the given pattern\n* <code>--match</code> - Filters by a field specified in `pyproject.toml`\n',
    'author': 'Or Levi',
    'author_email': 'orlevi128@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/orlevii/realm',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import os
from typing import List

from realm.entities import Project

from ..entities.context import RealmContext
from ..utils.child_process import ChildProcess


def apply_since_filters(ctx: RealmContext, since):
    changed_projects = get_changed_projects(ctx, since)
    dependent = set()
    for proj in ctx.projects:
        if is_project_affect_by_change(proj, changed_projects):
            dependent.add(proj)

    dependent = dependent.union(changed_projects)

    ctx.projects = dependent


def is_project_affect_by_change(project, changed_projects):
    changed_dependencies = [
        p
        for p
        in changed_projects
        if is_dependent(project, p)
    ]
    return any(changed_dependencies)


def is_dependent(project: Project, changed_project: Project):
    dependency = project.dependencies.get(changed_project.package_name)
    if dependency is None:
        return False
    if not isinstance(dependency, dict):
        # Path dependency is a dict
        return False
    dependency_path = dependency.get('path')
    if dependency_path is None:
        return False
    return project.source_dir.joinpath(dependency_path).resolve() == changed_project.source_dir


def get_changed_projects(ctx: RealmContext, since):
    changed_files = ChildProcess.run(f'git diff --name-only {since}')
    git_root = ChildProcess.run('git rev-parse --show-toplevel')
    git_root = git_root.strip() if git_root else ''

    relative_realm_repo_path = ctx.config.root_dir[len(git_root):].lstrip(os.sep)

    changed_files = parse_changed_files(changed_files)

    ctx.projects.sort(key=lambda p: len(str(p.source_dir)), reverse=True)
    changed = set()

    for file in changed_files:
        for proj in ctx.projects:
            project_path = os.path.join(relative_realm_repo_path,
                                        proj.relative_path)
            if file.startswith(project_path):
                changed.add(proj)
                continue
    return changed


def parse_changed_files(changed_files: str) -> List[str]:
    def handle_windows(p: str):
        return p.replace('/', os.sep)

    changed_files = changed_files if changed_files else ''
    return [
        handle_windows(f.strip())
        for f
        in changed_files.split('\n')
        if f
    ]
